# Convert (print to stdout) makeline to canonic form (i.e, normalize spaces) {{{
# Usage:  variable=$(mlp_make_canonic_mline  "  make line ")
mlp_make_canonic_mline () {
    local  mline="$1"  s
    while ((${#mline} > 0)); do
        s=${mline:0:1}
        case $s in
        [[:space:]] )   ;;
        ")"|"," )       echo -n "$s "   ;;
        ":"|"+"|"=" )   echo -n " $s "  ;;
        ";" )           echo -n "$s  "  ;;
        * )             echo -n "$s"    ;;
        esac
        mline=${mline:1}
    done
}
#}}}

# Preprocess Makefile to avoid repeated operations in 'mxp_match_makeline'. {{{
# Exit code:    0 -- OK,  2 -- some names in Makefile are syntactically incorrect.
# Variables set by this function:
# NOTE: associative arrays created by this function must be declared outside the function
# -- otherwise, in bash 4.1 they are local...
#   map : MXP_MAKEFILE-key --> regexp-to-match-nonparametrized-target-name
unset MXP_MAKEFILE_PATT1;   declare -A MXP_MAKEFILE_PATT1
#   map MXP_MAKEFILE-key --> space-separated-list-of-names-in-pattern
unset MXP_MAKEFILE_NAMES1;  declare -A MXP_MAKEFILE_NAMES1
#   map : MXP_MAKEFILE-key --> regexp-to-match-parametrized-target-name
unset MXP_MAKEFILE_PATT2;   declare -A MXP_MAKEFILE_PATT2
mlp_preporcess_makefile () {
    # Global variables set by this function:
    mlp_errors=()               # list of stings containing error messages, empty if OK
    
    report_mkre_error () {
        mlp_errors+=( "[ERR]  Invalid target name pattern in Makefile:  '$par_target'" )
        mlp_errors+=( "$(printf '%*s^\n' $((50+$i)) '')" )
    }
    
    local  par_target
    local  matched_targets=()   # list of matches in MXP_MAKEFILE -- it must be unique
    local  target_def           # the last substituted target definition (multiple is an error!)
    local  nc="[A-Za-z0-9]"  enc="[A-Za-z0-9_]"  xnc="[A-Za-z0-9_{}]"  lbr='\{'  rbr='\}'
    for par_target in ${!MXP_MAKEFILE[@]}; do
        # Convert par_target into regular expressions
        local  patt1="^"  patt2="^"  names=()
        local  state=0
        local  i  ic  cn
        for ((i=0; i<${#par_target}; i++)); do
            ic=${par_target:$i:1}
            case $state in
            0 ) if   [[ $ic =~ $enc ]]; then patt1+=$ic; patt2+=$ic       
                elif [[ $ic =~ $lbr ]]; then state=1; cn=""     
                else report_mkre_error; break; fi ;;
            1 ) if   [[ $ic =~ $nc ]];  then state=2; cn+=$ic   
                elif [[ $ic =~ $lbr ]]; then state=3            
                else report_mkre_error; break; fi ;;
            2 ) if   [[ $ic =~ $nc ]];  then cn+=$ic            
                elif [[ $ic =~ $rbr ]]; then state=0; patt1+="($nc+)"; names+=("{$cn}")
                                             patt2+="$nc*(\{($nc+)\}$nc*)*" 
                else report_mkre_error; break; fi ;;
            3 ) if   [[ $ic =~ $enc ]]; then cn+=$ic            
                elif [[ $ic =~ $rbr ]]; then state=4            
                else report_mkre_error; break; fi ;;
            4 ) if   [[ $ic =~ $rbr ]]; then state=0; patt1+="($enc*)"; names+=("{{$cn}}")
                                             patt2+="($xnc*)"
                else report_mkre_error; break; fi ;;
            esac
        done
        patt1+='$'  patt2+='$'
        if ((${#mlp_errors[@]} > 0)); then return 2; fi
        MXP_MAKEFILE_PATT1["$par_target"]="$patt1"
        MXP_MAKEFILE_NAMES1["$par_target"]="${names[@]}"
        MXP_MAKEFILE_PATT2["$par_target"]="$patt2"
    done
    return 0
}
#}}}

# Match makeline: scan MXP_MAKEFILE to find definition of a target and parse it. {{{
# Usage:  match_makeline  req-target        ???  makefile-name  "  make line "
#   req-target  -- the name of a target which definition has to be found
# Exit code:
#   0 -- success, global variables 'mlp_...' contain parsing results
#   1 -- error discovered, 'mlp_errors' contains list of error messages
#   2 -- errors in makefile, not relating to a particular target

mlp_match_makeline () {
    local req_target="$1"
    
    # Global variables set by this function:
    mlp_errors=()               # list of stings containing error messages, empty if OK
    mlp_orig_makeline=""        # original (parametrized) makeline, "target <= canonic"
    mlp_subst_makeline=""       # makeline after substitution (empty if orig is not parametrized)
    mlp_canonic_mline=""        # only definition (without target), substituted 
    # Other global variables are set by mlp_parse_makeline -- reset here for return on error
    mlp_required_targets=()     # sequence of names of required targets, without repetitions
    mlp_required_target_dirs=() # sequence of paths to required targets, without repetitions
    mlp_required_scripts=()     # sequence of required scripts, with extensions, without repetitions
    mlp_ops_sequence=()         # sequence of operation to perform, like "export+D=T", or "source+S"
    
    local  par_target
    local  matched_targets=()   # list of matches in MXP_MAKEFILE -- it must be unique
    local  target_def           # the last substituted target definition (multiple is an error!)
    local  nc="[A-Za-z0-9]"  enc="[A-Za-z0-9_]"  lbr='\{'  rbr='\}'
    # 1. Create list of all matching parametrized targets
    for par_target in ${!MXP_MAKEFILE[@]}; do
        # Try to match against required target
        if [[ $req_target =~ ${MXP_MAKEFILE_PATT1[$par_target]} ]]; then
            matched_targets+=("$par_target")
        fi
    done
    # 2. Find most specific matching target(s)
    mlp_cmp_partargets () {     # Returns: 0 -- uncomparable, 1 -- $1<=$2, 2 -- $1>=$2
                                # ('a<=b' means 'a is more specific than b')
        if [[ $1 =~ ${MXP_MAKEFILE_PATT2[$2]} ]]; then
            if [[ $2 =~ ${MXP_MAKEFILE_PATT2[$1]} ]]; then return 0; else return 1; fi
        else
            if [[ $2 =~ ${MXP_MAKEFILE_PATT2[$1]} ]]; then return 2; else return 0; fi
        fi
    }
    local min_matched_targets=() mtarget add_target keep_targets cmp_res
    for mtarget in "${matched_targets[@]}"; do
        add_target=1;  keep_targets=()
        for mmtarget in ${min_matched_targets[@]}; do
            mlp_cmp_partargets "$mtarget" "$mmtarget";  cmp_res=$?
            case $cmp_res in
            0 ) keep_targets+=( "$mmtarget" ) ;;
            1 ) ;;
            2 ) add_target=0; keep_targets+=( "$mmtarget" ) ;;
            esac
        done
        if ((add_target)); then min_matched_targets=("$mtarget"); else min_matched_targets=(); fi
        min_matched_targets+=( "${keep_targets[@]}" )
    done
    
    # degug:
    #echo "Matched targets:  ${matched_targets[@]}"
    #echo "Most specific:    ${min_matched_targets[@]}"
    
    # Check wheter exactly 1 match found
    if ((${#min_matched_targets[@]} == 0)); then
        mlp_errors+=( "[ERR]  Unknown target: '$req_target'" )
        return 2
    elif ((${#min_matched_targets[@]} > 1)); then
        mlp_errors+=( "[ERR]  Multiple matching targets for '$req_target':" )
        mlp_errors+=( "        $(join "   " ${min_matched_targets[@]})" )
        return 2
    else
        mt=$min_matched_targets
        [[ $req_target =~ ${MXP_MAKEFILE_PATT1[$mt]} ]]
        # do substitutions in target definition
        target_def="${MXP_MAKEFILE[$mt]}"
        names=( ${MXP_MAKEFILE_NAMES1[$mt]} )
        for ((i=0; i<${#names[@]}; i++)); do
            target_def=$(echo "$target_def" | sed -e "s|${names[$i]}|${BASH_REMATCH[$i+1]}|g")
        done
        # prepare everything else
        mlp_canonic_mline=$(mlp_make_canonic_mline  "${MXP_MAKEFILE[$mt]}")
        mlp_orig_makeline="$mt <= $mlp_canonic_mline"
        if [[ $mt =~ $lbr ]]; then
            mlp_canonic_mline=$(mlp_make_canonic_mline  "$target_def")
            mlp_subst_makeline="$req_target <= $mlp_canonic_mline"
        fi
    fi
    
    # Parse substituted makeline
    mlp_parse_makeline  "$target_def"  $req_target
}
#}}}

# Parse makeline. {{{
# Usage:  mlp_parse_makeline  "  make line "  target-name
#   "  make line "  -- as it is stored in Makefile
#   target-name     -- used only for building error message
# Exit code:
#   0 -- success, global variables 'mlp_...' contain parsing results
#   1 -- error discovered, 'mlp_errors' contains list of error messages
#
# Makeline BNF syntax:
#   makeline    ::=  makegroups
#   makegroups  ::=  makegroup  ( ';'  makegroup )*
#   makegroup   ::=  ( '(' req-targets ')' )?  ( params : )?  methods
#   req-targets ::=  req-target  ( ',' req-target )*
#   req-target  ::=  env-var-name = target-name
#   params      ::=  param-name  ( '+' param-name )*
#   methods     ::=  method-name  ( '+' method-name )*
#
mlp_parse_makeline () {
    # The following global variables are set by parser:
    # Note: each string in mlp_ops_sequence contains a single space -- we replace it by '+',
    # so we can then export array as a single space-delimited variable, and then replace
    # first occurences of '+' by space in order to obtain legal command.
    mlp_required_targets=()     # sequence of names of required targets, without repetitions
    mlp_required_target_dirs=() # sequence of paths to required targets, without repetitions
    mlp_required_scripts=()     # sequence of required scripts, with extensions, without repetitions
    mlp_ops_sequence=()         # sequence of operation to perform, like "export+D=T", or "source+S"
    
    local delims="(=,)+:;"
    local mline="$1"
    local target_name="$2"
    
    # Tokenize {{{
    local tokens=() token
    while (( ${#mline} > 0 )); do
        if  [[ "${mline:0:1}" =~ [[:space:]] ]]; then
            mline=${mline:1}
        elif [[ "${mline:0:1}" =~ [$delims] ]]; then
            tokens+=( "${mline:0:1}" )
            mline=${mline:1}
        else
            (( i = 0 ))
            while ((i < ${#mline})) && [[ ! ${mline:$i:1} =~ [[:space:]$delims] ]]; do ((i++)); done
            tokens+=( ${mline:0:$i} )
            mline=${mline:$i}
        fi
    done
    tokens+=( "<<<EOL" )
    #}}}
    
    # Parse: {{{
    # Parsing is performed by top-down method, folloving BNF syntax.
    
    # Supplementary functions {{{
    add_to_targets () {  # Args: $1 -- new-value {{{
        local  t  val=$1
        for t in ${mlp_required_targets[@]}; do if [ "$t" == "$val" ]; then return; fi; done
        mlp_required_targets+=( $val )
        mlp_required_target_dirs+=( $MXP_MAINDIR/$val )
    } #}}}
    add_to_scripts () {  # Args: $1 -- new-value {{{
        local  s  val=$1
        for s in ${mlp_required_scripts[@]}; do if [ "$s" == "$val" ]; then return; fi; done
        mlp_required_scripts+=( $val )
    } #}}}
    extend_pm_names () { # Args: $1 -- exension (sh or params.sh), $2,... -- simple script names {{{
        local ext=$1;  shift
        while (($# > 0)); do  echo $1.$ext;  shift;  done
    } #}}}
    local  curr_token  curr_token_idx=-1
    next_token () { (( curr_token_idx++ )); curr_token=${tokens[$curr_token_idx]}; }
    is_EOL   () { [ "$curr_token" == "<<<EOL" ]      &&  return 0  ||  return 1; }
    is_delim () { [[ "$curr_token" =~ ^[$delims]$ ]] &&  return 0  ||  return 1; }
    is_name  () { ! is_EOL  &&  ! is_delim; }
    report_error () {  # Args: $1 -- "expected..." message {{{
        local msg1="[ERR]  Definition of target '$target_name' is syntactically  incorrect:"
        local msg2="         $mlp_canonic_mline"
        local prev_tokens="$curr_token"  i=$((curr_token_idx - 1))
        while (( i >= 0 && i >= curr_token_idx-3 )); do
            prev_tokens="${tokens[$i]} $prev_tokens"
            (( i-- ))
        done
        if (( i > 0 )); then prev_tokens="... $prev_tokens"; fi
        local msg3="       -- at tocken $((curr_token_idx+1)): expected $1, got '$curr_token'"
        msg3+=" (last tokens: \"$prev_tokens\")."
        mlp_errors+=( "$msg1" "$msg2" "$msg3" )
    } #}}}
    #}}}
    
    # Parsing functions {{{
    process_method_names () { #{{{
        local method_names=( $(extend_pm_names  sh  ${script_names[@]}) )
        for method_name in ${method_names[@]}; do
            add_to_scripts  $method_name
            mlp_ops_sequence+=( "source+\$(mxp_locate_and_record_script+$method_name)" )
        done
    } #}}}
    process_param_names () { #{{{
        local param_names=( $(extend_pm_names  params.sh  ${script_names[@]}) )
        for param_name in ${param_names[@]}; do
            add_to_scripts  $param_name
            mlp_ops_sequence+=( "source+\$(mxp_locate_and_record_script+$param_name)" )
        done
    } #}}}
    local script_names  # parse_script_names stores result in this array
    parse_script_names () { #{{{
        script_names=()
        is_name  ||   { report_error "name"; return 1; }
        script_names+=( $curr_token )
        next_token
        while [ "$curr_token" == "+" ]; do
            next_token
            is_name  ||   { report_error "name"; return 1; }
            script_names+=( $curr_token )
            next_token
        done
    } #}}}
    parse_req_target () { #{{{
        is_name  ||  { report_error "name"; return 1; }
        local var_name=$curr_token
        next_token
        [ "$curr_token" == "=" ]  ||  { report_error "'='"; return 1; }
        next_token
        is_name  ||  { report_error "name"; return 1; }
        local target_name=$curr_token
        add_to_targets $target_name
        mlp_ops_sequence+=( "export+$var_name=$MXP_MAINDIR/$target_name" )
        next_token
        return 0
    } #}}}
    parse_req_targets () { #{{{
        parse_req_target  ||  return 1
        while [ "$curr_token" == "," ]; do
            next_token
            parse_req_target  ||  return 1
        done
        [ "$curr_token" == ")" ]  &&  { next_token; return 0; }
        report_error "',' or ')'"
        return 1
    } #}}}
    parse_maybe_req_targets () { #{{{
        if [ "$curr_token" == "(" ]; then
            next_token
            parse_req_targets  ||  return 1
        fi
        return 0
    } #}}}
    parse_makegroup () { #{{{
        parse_maybe_req_targets ||  return 1
        parse_script_names      ||  return 1
        if [ "$curr_token" == ":" ]; then
            process_param_names
            next_token
            parse_script_names  ||  return 1
        fi
        process_method_names
        return 0
    } #}}}
    parse_makegroups () { #{{{
        parse_makegroup   ||  return 1
        while [ "$curr_token" == ";" ]; do
            next_token
            parse_makegroup   ||  return 1
        done
        is_EOL  &&  return 0
        report_error "';' or EOL"
        return 1
    } #}}}
    #}}}
    #}}}
    next_token
    parse_makegroups  || return 1
} #}}}
